#!/usr/bin/perl -w

use strict;

use Test::More;
use Test::FormValidator qw(no_plan);
use Carp ();
$SIG{__WARN__} = \&Carp::cluck;
$SIG{__DIE__} = \&Carp::confess;

use lib '/Users/gwg/go/scratch/tools/';
my $installation_dir = '/Users/gwg/go/scratch/tools/';  # alter as necessary

$ENV{TMPL_VERBOSE} = 1;

use Data::Dumper;
$Data::Dumper::Sortkeys = 1;


run_tests();

exit(0);

sub run_tests {


	# prerequisites
	my @list = qw( GO::Tool::GenericGoTool
		CGI::Carp
		Data::FormValidator::Constraints
		GO::Object::Generic
		GO::TestSet
		GO::Utilities
		GO::GeneralPurposeParser
		URI::Escape
		GO::Tool::AddTool
		GO::Object::Tool
		GO::Object::ToolPreview
		GO::Object::Developer
	);
	foreach (@list)
	{	use_ok($_);
	}
	
	print STDERR "Tool dfv_profile: ".Dumper(GO::Object::Tool->dfv_profile);
	
	print STDERR "Tool Preview dfv_profile: ".Dumper(GO::Object::ToolPreview->dfv_profile);
	
	
	
	my $tfv = Test::FormValidator->new;
	my $tool = GO::Tool::AddTool->new;

	my $add_tool_spec = $tool->get_spec();
	## let's make an easy-to-access summary of this stuff...
	my $tool_spec_data;
	foreach my $param (keys %$add_tool_spec)
	{	foreach my $type qw(default required list_values)
		{	if ($add_tool_spec->{$param}{$type})
			{	# this param has a default...
				$tool_spec_data->{by_type}{$type}{$param} = $add_tool_spec->{$param}{$type};
				$tool_spec_data->{by_param}{$param}{$type} = $add_tool_spec->{$param}{$type};
			}
		}
	}
	
	$tfv->profile( $tool->dfv_profile );

	$tfv->prefix('AddTool form: ');

	$tfv->check_not_ok({}, "No data supplied");
	# get the results. Anything with a default value should be specified.
	my $results = $tfv->results();
	
	

	my $var_h = {
		name => 'test tool', 
		url => 'http://www.here.com', 
		email => 'me@home.com', 
		developer => 'LBL, http://www.lbl.gov', 
		tool_type => 'is_online_tool', 
		feature => 'ont_view', 
		description => 'Test tool is a wikkid tool that does loads of cool stuff.',
		is_open_source => 'true',
		license => 'free_academic',
		rm => 'preview',
	};

	$tfv->check_ok($var_h, "Hash, all valid");

	$var_h = {
		name => 'My Test Tool', 
		url => 'http://www.here.com', 
		email => 'me@home.com', 
		developer => 'LBL, http://www.lbl.gov', 
		tool_type => ['is_online_tool', 'is_standalone_tool'],
		compatible_os => 'mac',
		feature => ['ont_view', 'godb', 'stat'],
		license => 'free_academic',
		description => 'My Test Tool is a wikkid tool that does loads of cool stuff.',
		go_data_used => ['term','def','synonym','tree',],
		is_open_source => 'false',
	};
	$tfv->check_ok($var_h, 'Hash, should be valid');
	undef $results; $results = $tfv->results();
	print STDERR "results: ".Dumper($results);

=cut
	my $mech = Test::WWW::Mechanize::CGIApp->new;
	$mech->app("GO::Tool::AddTool");
	$mech->get_ok("", "Got GO::Tool::AddTool with no args");

	$mech->get_ok("?rm=tool_form");
	$mech->get_ok("?rm=blurb");
	$mech->get_ok("?name=blobby");
	$mech->submit_form_ok( { fields => { test_data => 'blah blah blah' } } , "Submitting crappy data with an rm" );
	$var_h = { name => 'test tool' };


	$var_h->{feature} = [ 'ont_view', 'annot_view' ];

	$var_h->{name} = [ 'Tool 1', 'Tool 2' ];

	$var_h->{name} = 'Tool 1';


	
=cut
}

sub make_url_from_hash {
	my $var_h = shift;
	my $url = join("&", 
		map { 
			my $k = $_;
			if (!ref $var_h->{$k})
			{	$k . "=" . $var_h->{$k}
			}
			else
			{	join "&", map { $k ."=". $_ } @{$var_h->{$k}}
			}
		} keys %$var_h);
	return $url;
}

=cut

args: spec: the abbreviated spec
      mech: the mech
      input: what was used as input, as a hash
    [ use_form: the number of the form to use ]

=cut
sub compare_inputs_to_submitted_values {
	my $args = shift;
	if (! $args->{spec} || ! $args->{mech} || ! $args->{input} )
	{	print STDERR "missing an important parameter!\n" && return undef;
	}

	my $mech_thing = $args->{mech};
	my @form_arr = $mech_thing->forms();

	my $n = $args->{use_form} || 0;
	print STDERR "No forms found!\n" && return undef unless $form_arr[$n];
	
	my @input_arr = $form_arr[$n]->inputs();
	
	print STDERR "No inputs found!\n" && return undef unless scalar @input_arr > 0;
	
	my $input_h;
	foreach (@input_arr)
	{	next if $_->type eq 'submit';
		next if $_->name eq 'rm';
		if ($_->type eq 'hidden')
		{	if ($input_h->{$_->name})
			{	if (! ref $input_h->{$_->name} )
				{	$input_h->{$_->name} = [ $input_h->{$_->name}, $_->value ];
				}
				else
				{	push @{$input_h->{$_->name}}, $_->value;
				}
			}
			else
			{	$input_h->{$_->name} = $_->value;
			}
		}
	}

	foreach (keys %{$args->{spec}{default}})
	{	next if $args->{input}{$_};
		## delete the ones that weren't in our original query
		delete $input_h->{$_};
	}
	print STDERR "input_h: ".Dumper($input_h);
	return $input_h;
}
